# pysim.py
#
# Derek Devnich             14 March 2007
'''This is a framework for a neural network gui, written in PyGame.'''

#-------------------------------------------------
# Setup
#-------------------------------------------------

import dataTypes
import pygame, sys, os  #, random
from pygame.locals import *
# random.seed(os.urandom(99))  now done in lisamain.py

# Set screen measurements
gui_x = 90                          # Node width
gui_y = 40                          # Node height
label_y = 16                        # Text height (also text point size)

# Initialize pygame
pygame.font.init()
symfont = pygame.font.SysFont('times', label_y)
clock = pygame.time.Clock()

# Model control variables
model = ['shell','run']             # command line/simulation run switch
sym_file = None                     # Symfile name
gui = True                          # gui switch
pause = False                       # pause switch
view = ['network','act','hebb']     # text view switch
fps = 30                            # frames per second
frame_ratio = 1.5                   # frame rate multiplier

# Define colors
BLACK = (0,0,0)
WHITE = (255,255,255)
RED = (255,0,0)
GREEN = (0,255,0)
YELLOW = (255,255,50)
ORANGE = (255,100,0)
BLUE = (50,150,255)
GRAY = (150,150,150)

# Drawing containers
update_list = []


#-------------------------------------------------
# Define GUI elements
#-------------------------------------------------

class Label(object):
    '''Defines a free-standing text field.'''

    def __init__(self, screen, name, x, y, update_flag=True, label_type='normal'):
        self.name = name
        self.label_type = label_type
        self.update_flag = update_flag          # Update label on next iteration

        # Define location of the label.  A rect is defined as:
        # (x coordinate, y coordinate, width, height)
        self.rect = (x, y, gui_x*2, label_y)

        # Render text      
        self.text = symfont.render(name, True, WHITE, BLACK)

    def move(self, screen, diff_x, diff_y, update_list):
        '''Changes location of a unit.  The whole screen needs to be
        updated when this happens, by calling NetworkDisplay.update_area().'''

        screen.fill(BLACK, self.rect)
        update_list.append(self.rect)

        self.rect = (self.rect[0] + diff_x, self.rect[1] + diff_y, self.rect[2], self.rect[3])
        screen.blit(self.text, self.rect)

        update_list.append(self.rect)
        
    def update_label(self, update_list, screen, label_type=None, new_label=None):
        '''Updates the color of the text based on its label_type,
        blits the text to the label's coordinates, and appends those
        coordinates to the update list.  This function gets called any
        time a label is added or changed.'''

        screen.fill(BLACK, self.rect)
        if new_label is not None:
            self.name = new_label
        if label_type is not None:
            self.label_type = label_type
        if self.label_type == 'normal':
            self.text = symfont.render(self.name, True, WHITE, BLACK)
        elif self.label_type == 'driver':
            self.text = symfont.render(self.name, True, YELLOW, BLACK)
        screen.blit(self.text, self.rect)
        update_list.append(self.rect)

    def update_location(self, screen, diff_y):
        '''Changes location of a unit.  The whole screen needs to be
        updated when this happens, by calling NetworkDisplay.update_area().'''

        self.rect = (self.rect[0], self.rect[1] + diff_y, self.rect[2], self.rect[3])
        self.update_flag = True


class NodeDisplay(object):
    '''Represents a single unit in the network.'''

    def __init__(self, screen, unit, color, x, y, update_flag=True, retrieved=False, offset=4):
        self.name = unit.name
        self.unit = unit
        self.color = color
        self.retrieved = retrieved
        self.act = 0.0
        self.input = 0.0
        self.update_flag = update_flag

        # Define locations of node and subfields
        self.rect = (x, y, gui_x, gui_y)
        self.text_rect = (x + offset, y + offset, gui_x - 10, 15)
        self.input_rect = (x + offset, self.text_rect[1] + self.text_rect[3] + offset, gui_x - 10, 3)
        self.act_rect = (x + offset, self.input_rect[1] + self.input_rect[3], gui_x - 10, 10)

        # Render content        
        self.border = pygame.draw.rect(screen, color, self.rect, 1)
        self.text = symfont.render(self.name, True, color, BLACK)

    def set_retrieved(self, screen, update_list): #, retrieved=None):
        '''Redraw border based on retrieved status of node.'''

        if type(self.unit) is dataTypes.GroupUnit or type(self.unit) is dataTypes.PropUnit or type(self.unit) is dataTypes.SPUnit or type(self.unit) is dataTypes.OPUnit:
            self.retrieved = self.unit.retrieved
        else:
            self.retrieved = False

        # Set border based on self.retrieved
        if self.retrieved == True:
            self.border = pygame.draw.rect(screen, WHITE, self.rect, 3)
        else:
            screen.fill(BLACK, self.rect)
            screen.blit(self.text, self.text_rect)
            self.border = pygame.draw.rect(screen, self.color, self.rect, 1)
        update_list.append(self.rect)

    def update_location(self, screen, diff_y):
        '''Changes location of a unit.  The whole screen needs to be
        updated when this happens, by calling NetworkDisplay.update_area().'''

        self.rect = (self.rect[0], self.rect[1] + diff_y, self.rect[2], self.rect[3])
        self.text_rect = (self.text_rect[0], self.text_rect[1] + diff_y, self.text_rect[2], self.text_rect[3])
        self.input_rect = (self.input_rect[0], self.input_rect[1] + diff_y, self.input_rect[2], self.input_rect[3])
        self.act_rect = (self.act_rect[0], self.act_rect[1] + diff_y, self.act_rect[2], self.act_rect[3])

        self.update_flag=True

    def move(self, screen, diff_x, diff_y, update_list):
        '''Changes location of a unit.  The whole screen needs to be
        updated when this happens, by calling NetworkDisplay.update_area().'''

        screen.fill(BLACK, self.rect)
        self.border = pygame.draw.rect(screen, BLACK, self.rect, 3)
        update_list.append(self.rect)

        self.rect = (self.rect[0] + diff_x, self.rect[1] + diff_y, self.rect[2], self.rect[3])
        self.text_rect = (self.text_rect[0] + diff_x, self.text_rect[1] + diff_y, self.text_rect[2], self.text_rect[3])
        self.input_rect = (self.input_rect[0] + diff_x, self.input_rect[1] + diff_y, self.input_rect[2], self.input_rect[3])
        self.act_rect = (self.act_rect[0] + diff_x, self.act_rect[1] + diff_y, self.act_rect[2], self.act_rect[3])
        screen.blit(self.text, self.text_rect)

        update_list.append(self.rect)

    def update_act(self, update_list, act=None):
        '''Updates the activation bar of a unit, based on its new or
        current activation.  This function needs to be called from a
        loop that iterates through all the units in the network.'''

        if act is None:
            act = self.unit.act

        # color of child prop
        if type(self.unit) == dataTypes.PropUnit:
            if self.unit.mode < 0:
                self.color = RED
            else:
                self.color = YELLOW

        screen.fill(BLACK, self.act_rect)
        current_act = (self.act_rect[0], self.act_rect[1], self.act_rect[2] * act, self.act_rect[3])
        screen.fill(self.color, current_act)
        update_list.append(self.act_rect)

    def update_input(self, update_list, source=None):
        '''Updates the input bar of a unit, based on its new or
        current input.  This function needs to be called from a loop
        that iterates through all the units in the network.

        This function will have to be scaled to deal with the various input
        sources.  Generally, source should be a number in the range [0:1].'''

        if source is None:
            source = self.input
        screen.fill(BLACK, self.input_rect)
        current_input = (self.input_rect[0], self.input_rect[1], self.input_rect[2] * source,
                         self.input_rect[3])
        screen.fill(GRAY, current_input)
        update_list.append(self.input_rect)


#-------------------------------------------------
# Define GUI control objects
#-------------------------------------------------

class NetworkDisplay(object):
    '''Defines a container that controls the location, drawing, and
    display of all elements of the graphical portion of the GUI.'''

    def __init__(self, y=0):
        # These lists contain all Label and NodeDisplay objects
        self.nodes = {'target':{'label':[],'group':[],'prop':[],'sp':[],'pred':[],'obj':[]},
                      'source':{'label':[],'group':[],'prop':[],'sp':[],'pred':[],'obj':[]},
                      'schema':{'label':[],'group':[],'prop':[],'sp':[],'pred':[],'obj':[]},
                      'ltm':{'label':[],'group':[],'prop':[],'sp':[],'pred':[],'obj':[]},
                      'semantic':{'label':[],'pred':[],'obj':[], 'group':[]},
                      'status':{'input':[],'speed':[],'pause':[], 'iteration':[], 'top-down OK':[]}
                      }

        # Keywords used to access self.nodes in the correct order
        self.kwds = [['target','label'],
                     ['target','group'],
                     ['target','prop'],
                     ['target','sp'],
                     ['target','pred'],
                     ['target','obj'],
                     ['source','label'],
                     ['source','group'],
                     ['source','prop'],
                     ['source','sp'],
                     ['source','pred'],
                     ['source','obj'],
                     ['schema','label'],
                     ['schema','group'],
                     ['schema','prop'],
                     ['schema','sp'],
                     ['schema','pred'],
                     ['schema','obj'],
                     ['ltm','label'],
                     ['ltm','group'],
                     ['ltm','prop'],
                     ['ltm','sp'],
                     ['ltm','pred'],
                     ['ltm','obj'],
                     ['semantic','label'],
                     ['semantic','pred'],
                     ['semantic','obj'],
                     ['semantic','group'],
                     ['status','input'],
                     ['status','speed'],
                     ['status','pause'],
                     ['status','iteration'],
                     ['status','top-down OK']
                     ]

        # Status field locations
        self.status = {'input':0,'speed':1,'pause':2,'iteration':3, 'top-down OK':4}

        # Possible sources of input
        self.input = ['net input','top-down','bottom-up','lateral','outprop','hebb','hebb weight','None']

    def update_area(self, screen):
        '''This function recalculates the position of all display
        elements.  It iterates through each list of units in
        self.nodes, positioning the first element of each list
        directly below the last row of the previous list.  If the
        previous list is empty, the function looks up through
        self.nodes until it finds a list isn't.'''

        for kwd in self.kwds:
            if self.kwds.index(kwd) > 0:
                prior = self.kwds[self.kwds.index(kwd)-1]
                all_empty = False
                while len(self.nodes[prior[0]][prior[1]]) == 0:
                    if self.kwds.index(prior) > 0:
                        prior = self.kwds[self.kwds.index(prior)-1]
                    else:
                        all_empty = True
                        break
                if all_empty == False:
                    try:
                        if self.nodes[kwd[0]][kwd[1]][0].rect[1] <= self.nodes[prior[0]][prior[1]][-1].rect[1]:
                            new_y = self.nodes[prior[0]][prior[1]][-1].rect[1] + gui_y
                            old_y = self.nodes[kwd[0]][kwd[1]][0].rect[1]
                            diff_y = new_y - old_y
                            for node in self.nodes[kwd[0]][kwd[1]]:
                                node.update_location(screen, diff_y)
                            screen.fill(BLACK)
                    except IndexError:
                        continue

    def add_status(self, screen, status_text, kwd=['status','input']):
        '''Adds a status text field to the bottom of the screen.'''

        height = pygame.display.get_surface().get_height()
        y = 0 # i.e., top; to put at bottom: height - label_y - 4
        x = 500 + self.status[kwd[1]] * gui_x
        self.nodes[kwd[0]][kwd[1]].append(Label(screen, kwd[1] + ': ' + status_text, x, y,
                                                update_flag=True, label_type='normal'))
        # print "Added status", status_text

    def add_label(self, screen, name, kwd=['target','label']):
        '''Adds a label to the screen.  This is used mainly for section titles.'''

        home = kwd
        all_empty = False
        while len(self.nodes[kwd[0]][kwd[1]]) == 0:
            if self.kwds.index(kwd) > 0:
                kwd = self.kwds[self.kwds.index(kwd)-1]
            else:
                all_empty = True
                break
        if all_empty == False:
            x = 0
            y = self.nodes[kwd[0]][kwd[1]][-1].rect[1] + gui_y
            self.nodes[home[0]][home[1]].append(Label(screen, name, x, y, update_flag=True, label_type='normal'))
        else:
            self.nodes[home[0]][home[1]].append(Label(screen, name, 0, 0, update_flag=True, label_type='normal'))

    def add_node(self, unit, screen, kwd=['target','group'], color=GRAY):
        '''Adds a DisplayNode to the screen.'''

        width = pygame.display.get_surface().get_width()

        # If the current list already has units, cue off of the last one
        try:
            (x,y,q,r) = self.nodes[kwd[0]][kwd[1]][-1].rect
            # Add the unit to the same row, if it will fit
            if width >= x + 2*gui_x:
                self.nodes[kwd[0]][kwd[1]].append(NodeDisplay(screen, unit, color, x+gui_x, y, update_flag=True))
            else:
                # Start new row, which requires repainting everything
                x = 0
                y = self.nodes[kwd[0]][kwd[1]][-1].rect[1] + gui_y
                self.nodes[kwd[0]][kwd[1]].append(NodeDisplay(screen, unit, color, x, y, update_flag=True))
                self.update_area(screen)

        # If the list is empty, go up the lists until you find one that isn't
        except IndexError:
            home = kwd
            all_empty = False
            while len(self.nodes[kwd[0]][kwd[1]]) == 0:
                if self.kwds.index(kwd) > 0:
                    kwd = self.kwds[self.kwds.index(kwd)-1]
                else:
                    all_empty = True
                    break
            if all_empty == False:
                x = 0
                y = self.nodes[kwd[0]][kwd[1]][-1].rect[1] + gui_y
                self.nodes[home[0]][home[1]].append(NodeDisplay(screen, unit, color, x, y, update_flag=True))
                self.update_area(screen)
            else:
                self.nodes[home[0]][home[1]].append(NodeDisplay(screen, unit, color, 0, 0, update_flag=True))

    def translate(self, screen, diff_x, diff_y, update_list):
        # moves everything left, right up or down
        for analog in self.nodes.itervalues():
            for nodes in analog.itervalues():
                for node in nodes:
                    node.move(screen, diff_x, diff_y, update_list)
        pygame.display.update(update_list)
        update_list = []

    def draw_network(self, update_list, flag_needed=False):
        '''Draws the complete state of the network to the screen.'''

        screen.fill(BLACK)
        for analog in master.nodes.itervalues():
            for nodes in analog.itervalues():
                for node in nodes:
                    if flag_needed == False or (flag_needed == True and node.update_flag == True):
                        if type(node) == Label:
                            screen.blit(node.text, node.rect)
                            update_list.append(node.rect)
                        elif type(node) == NodeDisplay:
                            screen.blit(node.text, node.text_rect)
                            node.set_retrieved(screen, update_list)
                            node.update_act(update_list, act=None)
                            node.update_input(update_list, source=None)
                        node.update_flag = False
        pygame.display.update(update_list)
        update_list = []

#-------------------------------------------------
# Create NetworkDisplay object
#-------------------------------------------------
master = NetworkDisplay()


# add_node(self, unit, screen, kwd=['target','group'], color=GRAY):
# add_label(self, screen, name, kwd=['target','label'])

def SetUnitLocations(analogs, pred_sem, obj_sem, group_sem):
    master.__init__() # init master
    # locate the analogs & token units
    for analog in analogs:
        # NOTE: The following ASSUMES that the first analog in the sym file is te target, the second is
        #       the source, the third the schema and the fourth is LTM
        if analog.index == 0:
            analog_label = 'target'
        elif analog.index == 1:
            analog_label = 'source'
        elif analog.index == 2:
            analog_label = 'schema'
        elif analog.inxex == 3:
            analog_label = 'ltm'
        # make the analog's label
        master.add_label(screen, analog.name, [analog_label,'label'])
        # now iterate through units in analog...
        for group in analog.group:
            master.add_node(group, screen, kwd=[analog_label,'group'], color=ORANGE)
        for prop in analog.prop:
            master.add_node(prop, screen, kwd=[analog_label,'prop'], color=YELLOW)
        for SP in analog.sp:
            master.add_node(SP, screen, kwd=[analog_label,'sp'], color=BLUE)
        for pred in analog.pred:
            master.add_node(pred, screen, kwd=[analog_label,'pred'], color=GREEN)
        for obj in analog.obj:
            master.add_node(obj, screen, kwd=[analog_label,'obj'], color=RED)

    # locate the semantics
    for sem in pred_sem:
        master.add_node(sem, screen, kwd=['semantic','pred'], color=WHITE)
    for sem in obj_sem:
        master.add_node(sem, screen, kwd=['semantic','obj'], color=WHITE)
    for sem in group_sem:
        master.add_node(sem, screen, kwd=['semantic','group'], color=WHITE)

    # go ahead and show the state of the network

    master.add_status(screen, master.input[0], kwd=['status','input'])
    master.add_status(screen, '', kwd=['status','speed'])
    master.add_status(screen, '', kwd=['status','pause'])
    master.add_status(screen, '', kwd=['status','iteration'])
    master.add_status(screen, '', kwd=['status','top-down OK'])
    master.draw_network(update_list, flag_needed=True)

class TextDisplay(object):
    '''Defines a container that draws the text view of the the network
    (i.e., current activation, hebb weights, and so on).'''

    def __init__(self):

        ### TEST CODE ###
        a = range(8)
        b = range(8)
        b.reverse()             # Note that these variables are called in self.act_values
        c = range(5)            # and self.hebb_values.
        d = range(5)
        d.reverse()
        ### END TEST CODE ###

        # Column titles for Activation display and Hebb display
        self.act_titles = ['Unit','Activation','Net Input','Bottom Up','Top Down','Lateral','Outprop','Hebb']
        self.hebb_titles = ['Unit','Hebbed Unit','Activation','Weight','Buffer']

        # Column values for Act and Hebb display.  Each unit is represented as a list.
        # The elements of each list correspond to the column titles listed above.
        self.act_values = {'group':[a,b],'prop':[],'rf':[],'pred':[],'obj':[]}
        self.hebb_values = {'group':[c,d],'prop':[],'rf':[],'pred':[],'obj':[]}

        # Traversal order for act_values and hebb_values
        self.order = ['group','prop','rf','pred','obj']

        # List of Label objects used to construct display
        self.act_display = []
        self.hebb_display = []

        # Used for sorting by different Hebb column titles
        self.hebb_toggle = [1,2,3,4]

    def add_act_text(self,screen):
        '''Adds text elements for Activation text display.'''

        self.act_display = []

        # Draw column titles
        x = 0
        y = 0
        for unit_label in self.act_titles:
            self.act_display.append(Label(screen, unit_label, x, y, update_flag=True, label_type='normal'))
            x += gui_x

        # Draw unit information
        x = 0
        y += label_y
        for unit_type in self.order:
            for unit in self.act_values[unit_type]:
                for item in unit:
                    self.act_display.append(Label(screen, str(item), x, y, update_flag=True, label_type='normal'))
                    x += gui_x
                x = 0
                y += label_y

    def add_hebb_text(self,screen):
        '''Adds text elements for Hebb text display.  The function
        checks the current sort criterion, moves it the front of all
        the unit lists in self.hebb_values, sorts the lists, and then
        moves it back to its home location.'''

        self.hebb_display = []

        # Draw column titles
        x = 0
        y = 0
        for label_index in xrange(len(self.hebb_titles)):
            if label_index == self.hebb_toggle[0]:
                self.hebb_display.append(Label(screen, self.hebb_titles[label_index], x, y,
                                               update_flag=True, label_type='driver'))
            else:
                self.hebb_display.append(Label(screen, self.hebb_titles[label_index], x, y,
                                              update_flag=True, label_type='normal'))
            x += gui_x

        # Draw unit information
        x = 0
        y += label_y
        for unit_type in self.order:
            for unit in self.hebb_values[unit_type]:
            # Move sort criterion to the front of the unit information list
                unit.insert(0, unit.pop(self.hebb_toggle[0]))
            self.hebb_values[unit_type].sort()
            # Move the sort criterion back
            for unit in self.hebb_values[unit_type]:
                unit.insert(self.hebb_toggle[0], unit.pop(0))

            for unit in self.hebb_values[unit_type]:
                for item in unit:
                    self.hebb_display.append(Label(screen, str(item), x, y, update_flag=True, label_type='normal'))
                    x += gui_x
                x = 0
                y += label_y

    def draw_text(self, update_list, screen, view):
        '''Draw text display to the screen.'''

        screen.fill(BLACK)
        pygame.display.flip()   # This is here because I can't guarantee that previous updates
                                # have been cleaned out of the update list
        if view[0] == 'act':
            for i in self.act_display:
                i.update_label(update_list, screen, label_type=None, new_label=None)
        elif view[0] == 'hebb':
            for i in self.hebb_display:
                i.update_label(update_list, screen, label_type=None, new_label=None)
        pygame.display.update(update_list)
        update_list = []

    def toggle_hebb_sort_order(self, update_list, screen, view):
        if view[0] == 'hebb':
            # Change sort criterion
            self.hebb_toggle.append(self.hebb_toggle.pop(0))

            # Redraw to screen
            self.add_hebb_text(screen)
            self.draw_text(update_list, screen, view)

#-------------------------------------------------
# Create the TextDisplay object
#-------------------------------------------------
text_display = TextDisplay()


# Initialize pygame screen
def_scr_width = 1500 # 1200 # default screen width
def_scr_height = 1000 # 750 # default screen height
screen_width = def_scr_width # 1280 at school; 1200 on laptop
screen_height = def_scr_height # 1024 at school;  750 on laptop

def setup_screen(screen_width, screen_height):
    screen = pygame.display.set_mode((screen_width, screen_height))#,FULLSCREEN)
    return screen

screen = setup_screen(screen_width, screen_height)

def update_display(simulation):         # simulation is a dictionary that contains all global variables

    # update list is the list of graphical objects to be updated: init it here
    update_list = []

    # no need to wait...
    # clock.tick(simulation['fps'])

    ########################################
    ### UPDATE GUI FOR CURRENT ITERATION ###
    ########################################

    # update the Graphical nodes
    for analog in master.nodes.itervalues():
        for node_list in analog.itervalues():
            for node in node_list:
                if type(node) is NodeDisplay:
                    node.set_retrieved(screen, update_list)
                    node.update_act(update_list, node.unit.act)

    # Update iteration label  YO!  How do we make it update ALL labels?
    for node in master.nodes['status']['iteration']:
        node.update_label(update_list, screen, label_type='normal',
                          new_label='iteration: ' + str(simulation['iteration']))
    for node in master.nodes['status']['top-down OK']:
        node.update_label(update_list, screen, label_type='normal',
                          new_label='top-down OK: ' + str(simulation['top_down_ok']))

    # make the changes to the screen
    pygame.display.update(update_list)
    update_list = []


    ############################
    ### CHECK FOR USER INPUT ###
    ############################

    # While paused, check for un-pause and other keystrokes
    if simulation['pause']:
        wait = pygame.event.wait()
        if not hasattr(wait,'key'):
            pass
        # Check for un-pause
        elif wait.key == K_s and wait.type == KEYDOWN and view[0] == 'network':
            simulation['pause'] = False
            for node in master.nodes['status']['pause']:
                node.update_label(update_list, screen, label_type='normal',
                                  new_label='pause: '+ str(simulation['pause']))
        # translate left, right up or down
        elif wait.key == K_RIGHT and wait.type == KEYDOWN:
            # translate left
            master.translate(screen, -10, 0, update_list)
        elif wait.key == K_LEFT and wait.type == KEYDOWN:
            # translate right
            master.translate(screen, 10, 0, update_list)
        elif wait.key == K_UP and wait.type == KEYDOWN:
            # translate down
            master.translate(screen, 0, 10, update_list)
        elif wait.key == K_DOWN and wait.type == KEYDOWN:
            # translate up
            master.translate(screen, 0, -10, update_list)

        # Toggle network and text view
        elif wait.key == K_DELETE and wait.type == KEYDOWN:
            view.append(view.pop(0))
            if view[0] == 'act':
                # Paint activation text to screen
                text_display.add_act_text(screen)
                text_display.draw_text(update_list, screen, view)
            elif view[0] == 'hebb':
                # Paint hebb text to the screen
                text_display.add_hebb_text(screen)
                text_display.draw_text(update_list, screen, view)
            elif view[0] == 'network':
                # Paint network to screen
                master.draw_network(update_list, flag_needed=False)
        # Toggle hebb sort view
        elif wait.key == K_DOWN and wait.type == KEYDOWN and view[0] == 'hebb':
            text_display.toggle_hebb_sort_order(update_list, screen, view)
        # Exit
        elif wait.key == K_ESCAPE:
            sys.exit(0)
        elif wait.key == K_q:
            simulation['simulation_complete'] = True
        # increase/decrease graphics refresh rate
        elif wait.key == K_EQUALS:  # increase speed
            simulation['gr_refresh_rate'] *= 2
            # DIAG
            print 'refresh rate now =',simulation['gr_refresh_rate']
        elif wait.key == K_MINUS:  # decrease speed
            simulation['gr_refresh_rate'] /= 2
            if simulation['gr_refresh_rate'] < 1:
                simulation['gr_refresh_rate'] = 1
            # DIAG
            print 'refresh rate now =',simulation['gr_refresh_rate']            

    # Get user input while un-paused
    for event in pygame.event.get():
        if not hasattr(event,'key'):
            continue
        elif event.key == K_s and event.type == KEYDOWN:
            if simulation['pause'] == False:
                simulation['pause'] = True
                for node in master.nodes['status']['pause']:
                    node.update_label(update_list, screen, label_type='normal',
                                      new_label='pause: ' + str(pause))
        # Exit
        elif event.key == K_ESCAPE:
            sys.exit(0)

        # translate left, right up or down
        elif event.key == K_RIGHT and event.type == KEYDOWN:
            # translate left
            master.translate(screen, -10, 0, update_list)
        elif event.key == K_LEFT and event.type == KEYDOWN:
            # translate right
            master.translate(screen, 10, 0, update_list)
        elif event.key == K_UP and event.type == KEYDOWN:
            # translate down
            master.translate(screen, 0, 10, update_list)
        elif event.key == K_DOWN and event.type == KEYDOWN:
            # translate up
            master.translate(screen, 0, -10, update_list)

        # Exit to LISA menu: NOTE: This does Not work (4/20/07)!  Only way to quit for now seems to be to
        #   hit Esc and quite Python as a whole...
        elif event.key == K_q:
            print '.'
            print 'USER HAS REQUESTED THAT I QUIT.'
            print '.'
            simulation['terminate_simulation'] = True # quit out of the whole simulation
            pygame.event.clear() # clears the event queue

        # increase/decrease graphics refresh rate
        elif event.key == K_EQUALS:  # increase speed
            simulation['gr_refresh_rate'] *= 2
        elif event.key == K_MINUS:  # decrease speed
            simulation['gr_refresh_rate'] /= 2
            if simulation['gr_refresh_rate'] < 1:
                simulation['gr_refresh_rate'] = 1
